from __future__ import annotations

import re
from functools import cached_property

from .unit import Unit
from .unit_registry import UnitRegistry
from . import unit_registry as reg
from .util import camel_to_snake


class Quantity:
    """
    Class used to represent a Quantity

    A quantity has a value and a unit
    """

    # unit_registry: UnitRegistry = None

    def __init__(self, value, unit: str | Unit = None):
        if isinstance(value, str):
            match = re.match(r"(?P<v>[+-]?((\d+\.\d*)|(\.\d+)|(\d+))([eE][+-]?\d+)?)(?P<u>.*)", value.strip())
            if match:
                res_dict = match.groupdict()
                self.value = float(res_dict['v'])
                if not unit:
                    self.unit = res_dict['u'].strip()
                else:
                    self.unit = unit
        else:
            self.value = value
            self.unit = unit

    def to(self, unit: str | Unit):
        if self.value is None:
            return None
        unit = self.unit_registry.get_unit(str(unit))
        if unit == self.unit:
            return self
        ref_value = self._unit.to_base_unit(self.value)
        value = unit.from_base_unit(ref_value)
        return self.__class__(value, unit)

    @classmethod
    def get_unit_registry(cls) -> UnitRegistry:
        return getattr(reg, camel_to_snake(cls.__name__))

    @cached_property
    def unit_registry(self):
        return self.get_unit_registry()

    @property
    def base_unit(self) -> Unit:
        return self.unit_registry.base_unit

    @property
    def units(self) -> list[Unit]:
        return self.unit_registry.units

    def to_base(self):
        return self.to(self.base_unit)

    def __repr__(self):
        if isinstance(self.value, float):
            return f"<{self.__class__.__name__}({self.value:.9}, '{self.unit.symbol}')>"
        return f"<{self.__class__.__name__}({self.value}, '{self.unit.symbol}')>"

    def __str__(self):
        return f'{self.value}'

    def __format__(self, format_spec=''):
        if (pos := format_spec.find('u')) > -1:
            unit = format(self.unit, format_spec[pos + 1:])
            format_spec = format_spec[0:pos]
            if unit:
                return ' '.join([format(self.value, format_spec), unit])

        return format(self.value, format_spec)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return self.__class__(self.value * other, self.unit)

        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.to_base().value == other.to_base().value

        return False

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return self.to_base().value > other.to_base().value

    def __ge__(self, other):
        if isinstance(other, self.__class__):
            return self.to_base().value >= other.to_base().value

    @property
    def unit(self) -> Unit:
        return self._unit

    @unit.setter
    def unit(self, value):
        if not value:
            self._unit = self.unit_registry.base_unit
        else:
            self._unit = self.unit_registry.get_unit(str(value))


class Length(Quantity):
    pass


class Area(Quantity):
    pass


class Volume(Quantity):
    pass


class Time(Quantity):
    pass


class Mass(Quantity):
    pass


class Force(Quantity):
    pass


class Substance(Quantity):
    pass


class Energy(Quantity):
    pass


class Velocity(Quantity):
    pass


class Temperature(Quantity):
    pass


class DeltaTemperature(Quantity):
    pass


class Pressure(Quantity):
    pass


class VolumeFlow(Quantity):
    pass


class MassDensity(Quantity):
    pass


class HeatFlow(Quantity):
    pass


class MolarFlow(Quantity):
    pass


class MassFlow(Quantity):
    pass


class MolarDensity(Quantity):
    pass


class MolarHeatCapacity(Quantity):
    pass


class MolarEntropy(Quantity):
    pass


class MolarHeat(Quantity):
    pass


class ThermalConductivity(Quantity):
    pass


class Viscosity(Quantity):
    pass


class SurfaceTension(Quantity):
    pass


class MassHeatCapacity(Quantity):
    pass


class MassEntropy(Quantity):
    pass


class MassHeat(Quantity):
    pass


class StandardGasFlow(Quantity):
    pass


class KinematicViscosity(Quantity):
    pass


class MolarVolume(Quantity):
    pass


class Fraction(Quantity):
    pass


class Dimensionless(Quantity):
    pass



from __future__ import annotations
from collections import namedtuple

from z_unit.constant import ATM

STP = namedtuple('STP', ['temperature', 'pressure'])


class Environment:
    """Global environment parameters"""
    def __init__(self):
        self.reset()
    
    def reset(self):
        """Reset environment parameters"""
        self._atmospheric_pressure = ATM
        self._standard_temperature = 293.15 # 20 degC
        self._standard_pressure = ATM

    @property
    def atmospheric_pressure(self) -> float:
        """Atmospheric pressure in Pa"""
        return self._atmospheric_pressure

    @atmospheric_pressure.setter
    def atmospheric_pressure(self, value: float):
        """Atmospheric pressure in Pa"""
        if value <= 0:
            raise ValueError("Atmospheric pressure shall be positive")
        self._atmospheric_pressure = value

    @property
    def stp(self) -> STP:
        """Standard temperature in K and pressure in Pa"""
        return STP(self.standard_temperature, self.standard_pressure)
    
    @property
    def standard_condition(self) -> STP:
        return self.stp
    
    @property
    def standard_temperature(self) -> float:
        """Standard temperature in K"""
        return self._standard_temperature
    
    @standard_temperature.setter
    def standard_temperature(self, value: float):
        """Standard temperature in K"""
        if value <= 0:
            raise ValueError("Standard temperature shall be positive")
        self._standard_temperature = value
    
    @property
    def standard_pressure(self) -> float:
        """Standard pressure in Pa"""
        return self._standard_pressure
    
    @standard_pressure.setter
    def standard_pressure(self, value: float):
        """Standard pressure in Pa"""
        if value <= 0:
            raise ValueError("Standard pressure shall be positive")
        self._standard_pressure = value
    
    def set_stp(self, *, temperature: float = 273.15, pressure: float = ATM):
        """Set standard temperature in K and pressure in Pa"""
        self.standard_temperature = temperature
        self.standard_pressure = pressure


# Global environment instance
_default_env = Environment()

def get_env() -> Environment:
    """Get the global environment instance"""
    return _default_env


import re


def multi_replace(string: str, replacements: dict) -> str:
    """
    Given a string and a replacement map, it returns the replaced string.

    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :rtype: str

    """
    # https://stackoverflow.com/a/36620263
    # Place longer ones first to keep shorter substrings from matching
    # where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against
    # the string 'hey abc', it should produce 'hey ABC' and not 'hey ABc'
    sub_strs = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile('|'.join(map(re.escape, sub_strs)))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


def camel_to_snake(s):
    return ''.join(['_'+c.lower() if c.isupper() else c for c in s]).lstrip('_')



from z_unit import unit


def test_unit():
    print("")
    print(repr(unit.kilojoule_per_kilogram_celsius))
    print("Format Test")
    print(f"Default Style: {unit.kilojoule_per_kilogram_celsius}")
    print(f"Quick Style: {unit.kilojoule_per_kilogram_celsius:q}")
    print(f"Python Style: {unit.kilojoule_per_kilogram_celsius:p}")
    print("multiply")
    print((unit.minute * unit.minute).factor)
    print((10 * unit.minute).factor)
    # assert False



from __future__ import annotations
from enum import Enum
from typing import Optional, Union, Callable, Iterable

from z_unit.environment import get_env
from z_unit.util import multi_replace
from z_unit import constant


class Unit:
    """
    Class used to represent a Unit.

    For unit conversion, units belong to same 'family' (unit set)
    shall be converted from/to one 'base unit' as reference.

    Creation: unit = Unit(name[, factor, offset])

    name: str
        the quick style name(symbol) of the unit,
        example: 'kg/s', 'm3/s', 'kJ/mol-C'

        Simple form used to represent definition, as:
        'm3' = 'm**3', 'mol-C' = 'mol*C'

    factor, offset: numeric value or function,
        for converting this unit to base unit.

        unit * factor + offset = base unit

    If unit conversion is nonlinear, to_base_unit() & from_base_unit()
    have to be overridden.

    Example
    -------
    >>> kilometer = Unit('km', factor=1e3)

    Which meter is base unit

    Properties
    ----------
    symbol: str
        the quick style name(symbol) of the unit.

    factor: float
        factor value

    offset: float
        offset value
    """

    def __init__(self, symbol: str, defined_by=None, factor: Union[float, Callable] = 1,
                 offset: Union[float, Callable] = 0, aliases: Iterable[str] = []):
        if isinstance(defined_by, Unit):
            factor = defined_by.factor

        if factor == 0:
            raise ValueError("Factor shall not be 0")

        self._symbol = symbol.replace(' ', '')
        self._factor = factor
        self._offset = offset
        self.aliases = aliases

    def to_base_unit(self, value: float):
        """
        Convert value from this unit to base unit
        :param value: numerical value
        :return: converted value
        """
        return self.factor * value + self.offset

    def from_base_unit(self, value: float):
        """
        Convert value from base unit to this unit
        :param value: numerical valve
        :return: converted valve
        """
        return (value - self.offset) / self.factor

    @property
    def symbol_quick_style(self):
        """
        The default style, quick & easy to typing

        Quick Style Convention:

        * No space separator
        * Power:    x ** y -> xy
        * Multiply: x * y  -> x-y
        * Parentheses will be omitted:  x / (y * z) -> x/y-z

        Examples:

            m ** 2 -> m2
            m * s -> m-s
            kJ / (C * kg) -> kJ/C-kg

        :return: quick style symbol in str
        """
        if self._symbol:
            return multi_replace(self._symbol, {
                '**': '',
                '*': '-',
                '(': '',
                ')': '',
            })

        return self._symbol

    @property
    def symbol_python_style(self):
        """
        Defined Style shows formula
        :return: defined style symbol in str
        """
        return self._symbol

    @property
    def symbol(self):
        return self.symbol_quick_style

    @property
    def factor(self):
        if callable(self._factor):
            return self._factor()

        return self._factor

    @property
    def offset(self):
        if callable(self._offset):
            return self._offset()

        return self._offset

    @property
    def aliases(self) -> list[str]:
        return self._aliases

    @aliases.setter
    def aliases(self, alist: Iterable[str]):
        self._aliases = list(set(alist).union([self.symbol]))

    def __repr__(self):
        return f"<Unit('{self}')>"

    def __str__(self):
        return self.symbol

    def __format__(self, format_spec=''):
        if format_spec == '':
            return self.symbol
        if format_spec == 'q':
            return self.symbol_quick_style
        if format_spec == 'p':
            return self.symbol_python_style

        raise ValueError('Invalid format specifier')

    def __mul__(self, other):
        # only for factor definition
        if isinstance(other, Unit):
            factor = self.factor * other.factor
            symbol = f"{self.symbol_python_style}*{other.symbol_python_style}"
            return Unit(symbol, factor=factor)

        if isinstance(other, (int, float)):
            factor = self.factor * other
            return Unit('dummy', factor=factor)

        return NotImplemented

    def __rmul__(self, other):
        factor = other * self.factor
        return Unit('dummy', factor=factor)

    def __truediv__(self, other):
        if isinstance(other, Unit):
            symbol = f"{self.symbol_python_style}/{other.symbol_python_style}"
            factor = self.factor / other.factor
            return Unit(symbol, factor=factor)

        if isinstance(other, (int, float)):
            factor = self.factor / other
            return Unit('dummy', factor=factor)

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, (int, float)):
            factor = other / self.factor
            return Unit('dummy', factor=factor)

        return NotImplemented

    def __pow__(self, power, modulo=None):
        if isinstance(power, (int, float)):
            symbol = f"{self.symbol_python_style}**{power}"
            factor = self.factor ** power
            return Unit(symbol, factor=factor)

        return NotImplemented


# basic unit
# length, m
meter = Unit('m')
kilometer = Unit('km', defined_by=1e3 * meter)
decimeter = Unit('dm', defined_by=1e-1 * meter)
centimeter = Unit('cm', defined_by=1e-2 * meter)
millimeter = Unit('mm', defined_by=1e-3 * meter)
micrometer = Unit('um', defined_by=1e-6 * meter)
foot = Unit('ft', factor=3.048e-1)
inch = Unit('in', factor=2.54e-2)

# area, m2
square_meter = Unit('m**2')
square_kilometer = Unit('km**2', defined_by=kilometer**2)
square_decimeter = Unit('dm**2', defined_by=decimeter**2)
square_centimeter = Unit('cm**2', defined_by=centimeter**2)
square_millimeter = Unit('mm**2', defined_by=millimeter**2)
square_micrometer = Unit('um**2', defined_by=micrometer**2)
square_foot = Unit('ft**2', defined_by=foot**2)
square_inch = Unit('in**2', defined_by=inch**2)

# volume, m3
cubic_meter = Unit('m**3')
cubic_centimeter = Unit('cm**3', defined_by=centimeter**3)
cubic_millimeter = Unit('mm**3', defined_by=millimeter**3)
liter = Unit('L', defined_by=decimeter**3)
milliliter = Unit('mL', defined_by=centimeter**3)
cubic_foot = Unit('ft**3', defined_by=foot**3)
cubic_inch = Unit('in**3', defined_by=inch**3)
# US_gallon
gallon = Unit('gal', defined_by=231 * cubic_inch)
barrel = Unit('bbl', defined_by=42 * gallon)

# time, s
second = Unit('s')
minute = Unit('min', defined_by=60 * second)
hour = Unit('hr', defined_by=60 * minute)
day = Unit('day', defined_by=24 * hour)
week = Unit('week', defined_by=7 * day)
year = Unit('yr', defined_by=365 * day)
month = Unit('mon', defined_by=year / 12)

# velocity, m/s
meter_per_second = Unit('m/s')
meter_per_minute = Unit('m/min', defined_by=meter / minute)
meter_per_hour = Unit('m/hr', defined_by=meter / hour)
kilometer_per_hour = Unit('km/hr', defined_by=kilometer / hour)
centimeter_per_second = Unit('cm/s', defined_by=centimeter / second)
foot_per_second = Unit('ft/s', defined_by=foot / second)
foot_per_minute = Unit('ft/min', defined_by=foot / minute)
foot_per_hour = Unit('ft/hr', defined_by=foot / hour)

# temperature base unit
celsius = Unit('C')
kelvin = Unit('K', offset=-273.15)
rankine = Unit('R', factor=5 / 9, offset=-273.15)
fahrenheit = Unit('F', factor=5 / 9, offset=-32 * 5 / 9)

# mass, kg
kilogram = Unit('kg')
gram = Unit('g', defined_by=1e-3 * kilogram)
tonne = Unit('t', defined_by=1e3 * kilogram)
pound = Unit('lb', defined_by=0.45359237 * kilogram)

# force, N
newton = Unit('N')
kilogram_meter_per_second_squared = Unit('kg*m/s**2', defined_by=newton)
kilo_newton = Unit('kN', defined_by=1e3 * newton)
dyne = Unit('dyn', defined_by=1e-5 * newton)
kilogram_force = Unit('kgf', defined_by=constant.G * kilogram)
tonne_force = Unit('tonf', defined_by=constant.G * tonne)
pound_force = Unit('lbf', defined_by=constant.G * pound)

# substance, kmol
kilomole = Unit('kmol')
mole = Unit('mol', defined_by=1e-3 * kilomole)
normal_cubic_meter = Unit('Nm**3', factor=1 / 22.414)
# @20 degC
T_60F = kelvin.from_base_unit(fahrenheit.to_base_unit(60))
T_0C = kelvin.from_base_unit(0)
T_20C = kelvin.from_base_unit(20)
standard_cubic_meter = Unit('Sm**3',
                            factor=lambda: normal_cubic_meter.factor * T_0C / get_env().stp.temperature)
standard_cubic_meter_20C = Unit('Sm**3_20C', defined_by=normal_cubic_meter * T_0C / T_20C)
standard_cubic_meter_60F = Unit('Sm**3_60F', defined_by=normal_cubic_meter * T_0C / T_60F)
# scf is @60 degF
standard_cubic_foot = Unit('SCF', defined_by=normal_cubic_meter * cubic_foot * T_0C / T_60F)
kilo_standard_cubic_foot = Unit('MSCF', defined_by=1e3 * standard_cubic_foot)
million_standard_cubic_foot = Unit('MMSCF', defined_by=1e6 * standard_cubic_foot)

# energy, kJ
kilojoule = Unit('kJ')
joule = Unit('J', defined_by=1e-3 * kilojoule)
megajoule = Unit('MJ', defined_by=1e6 * joule)
gigajoule = Unit('GJ', defined_by=1e9 * joule)
kilowatt_hour = Unit('kW*h', defined_by=kilojoule * hour)
kilowatt_year = Unit('kW*yr', defined_by=kilojoule * year)
calorie = Unit('cal', factor=4.184e-3)
kilocalorie = Unit('kcal', defined_by=1e3 * calorie)
megacalorie = Unit('Mcal', defined_by=1e6 * calorie)
gigacalorie = Unit('Gcal', defined_by=1e9 * calorie)
million_kilocalorie = Unit('MMkcal', defined_by=1e6 * kilocalorie)
british_thermal_unit = Unit('Btu', factor=1.055056)
million_british_thermal_unit = Unit('MMBtu', defined_by=1e6 * british_thermal_unit)
pound_force_foot = Unit('lbf*ft', defined_by=1e-3 * pound_force * foot)

# delta temperature
delta_celsius = Unit('C')
delta_kelvin = Unit('K', factor=1)
delta_rankine = Unit('R', factor=5 / 9)
delta_fahrenheit = Unit('F', factor=5 / 9)

class PressureType(Enum):
    ABSOLUTE = "absolute"
    GAUGE = "gauge"


class PressureUnit(Unit):
    def __init__(self, 
                 symbol: str,
                 defined_by: Optional[Union[str, Callable]] = None,
                 factor: float = 1.0,
                 offset: float = 0.0,
                 aliases: Iterable[str] = [],
                 pressure_type: PressureType = PressureType.ABSOLUTE):
        super().__init__(symbol=symbol, defined_by=defined_by, factor=factor, offset=offset, aliases=aliases)
        self.pressure_type = pressure_type

    @property
    def is_gauge(self):
        return not self.is_absolute

    @property
    def is_absolute(self):
        return self.pressure_type == PressureType.ABSOLUTE

    def to_base_unit(self, value: float) -> float:
        """convert to base unit value"""
        # if self.is_base_unit:
        # return value

        if self.is_gauge:
            # convert to absolute pressure
            abs_value = value * self.factor
            return abs_value + get_env().atmospheric_pressure

        return self.factor * value + self.offset

    def from_base_unit(self, value: float) -> float:
        """convert from base unit value"""
        # if self.is_base_unit:
        #     return value

        converted = (value - self.offset) / self.factor

        if self.is_gauge:
            # convert to gauge pressure
            guage_value = value - get_env().atmospheric_pressure
            return guage_value / self.factor

        return converted


# pressure base unit
pascal = PressureUnit('Pa')
kilopascal = PressureUnit('kPa', defined_by=1e3 * pascal)
megapascal = PressureUnit('MPa', defined_by=1e6 * pascal)
bar = PressureUnit('bar', defined_by=1e5 * pascal)
millibar = PressureUnit('mbar', defined_by=1e-3 * bar)
atm = PressureUnit('atm', factor=constant.ATM)
kg_force_per_square_centimeter = PressureUnit('kgf/cm**2', defined_by=kilogram_force / square_centimeter)
psi = PressureUnit('psi', defined_by=pound_force / square_inch)
pound_force_per_square_foot = PressureUnit('lbf/ft**2', defined_by=pound_force / square_foot)
torr = PressureUnit('torr', defined_by=atm / 760)
mm_Hg = PressureUnit('mmHg_0C', defined_by=torr)
inch_Hg = PressureUnit('inHg_32F', factor=1e3 * 3.386389)
inch_Hg_60F = PressureUnit('inHg_60F', factor=1e3 * 3.37685)
kilopascal_gauge = PressureUnit('kPag', factor=kilopascal.factor, pressure_type=PressureType.GAUGE)
megapascal_gauge = PressureUnit('MPag', factor=megapascal.factor, pressure_type=PressureType.GAUGE)
bar_gauge = PressureUnit('barg', factor=bar.factor, pressure_type=PressureType.GAUGE)
millibar_gauge = PressureUnit('mbarg', factor=millibar.factor, pressure_type=PressureType.GAUGE)
kg_force_per_square_centimeter_gauge = PressureUnit('kgf/cm**2_g', factor=kg_force_per_square_centimeter.factor,
                                            pressure_type=PressureType.GAUGE)
psi_gauge = PressureUnit('psig', factor=psi.factor, pressure_type=PressureType.GAUGE)
pound_force_per_square_foot_gauge = PressureUnit('lbf/ft**2_g', factor=pound_force_per_square_foot.factor,
                                         pressure_type=PressureType.GAUGE)
torr_gauge = PressureUnit('torr_g', factor=torr.factor, pressure_type=PressureType.GAUGE)
mm_Hg_gauge = PressureUnit('mmHg_0C_g', factor=mm_Hg.factor, pressure_type=PressureType.GAUGE)
inch_Hg_gauge = PressureUnit('inHg_32F_g', factor=inch_Hg.factor, pressure_type=PressureType.GAUGE)
inch_Hg_60F_gauge = PressureUnit('inHg_60F_g', factor=inch_Hg_60F.factor, pressure_type=PressureType.GAUGE)

# molar flow, base: kmol/s
kilomole_per_second = Unit('kmol/s')
kilomole_per_hour = Unit('kmol/h', defined_by=kilomole / hour)
kilomole_per_minute = Unit('kmol/min', defined_by=kilomole / minute)
normal_cubic_meter_per_hour = Unit('Nm**3/h', defined_by=normal_cubic_meter / hour)
normal_cubic_meter_per_day = Unit('Nm**3/d', defined_by=normal_cubic_meter / day)
# standard_cubic_meter_per_hour = Unit('Sm**3/h', defined_by=standard_cubic_meter / hour)
standard_cubic_meter_per_hour = Unit(
    'Sm**3/h',
    factor=lambda: normal_cubic_meter.factor * T_0C / get_env().stp.temperature / hour.factor)
standard_cubic_meter_20C_per_hour = Unit('Sm**3_20C/h', defined_by=standard_cubic_meter_20C / hour)
standard_cubic_meter_60F_per_hour = Unit('Sm**3_60F/h', defined_by=standard_cubic_meter_60F / hour)
# standard_cubic_meter_per_day = Unit('Sm**3/d', defined_by=standard_cubic_meter / day)
standard_cubic_meter_per_day = Unit(
    'Sm**3/d',
    factor=lambda: normal_cubic_meter.factor * T_0C / get_env().stp.temperature / day.factor)
standard_cubic_meter_20C_per_day = Unit('Sm**3_20C/d', defined_by=standard_cubic_meter_20C / day)
standard_cubic_meter_60F_per_day = Unit('Sm**3_60F/d', defined_by=standard_cubic_meter_60F / day)
mole_per_hour = Unit('mol/h', defined_by=mole / hour)
mole_per_minute = Unit('mol/min', defined_by=mole / minute)
mole_per_second = Unit('mol/s', defined_by=mole)
standard_cubic_foot_per_day = Unit('SCFD', defined_by=standard_cubic_foot / day)
kilo_standard_cubic_foot_per_hour = Unit('MSCFH', defined_by=kilo_standard_cubic_foot / hour)
kilo_standard_cubic_foot_per_day = Unit('MSCFD', defined_by=kilo_standard_cubic_foot / day)
million_standard_cubic_foot_per_hour = Unit('MMSCFH', defined_by=million_standard_cubic_foot / hour)
million_standard_cubic_foot_per_day = Unit('MMSCFD', defined_by=million_standard_cubic_foot / day)

# mass flow, base: kg/s
kilogram_per_second = Unit('kg/s')
kilogram_per_hour = Unit('kg/h', defined_by=kilogram / hour)
kilogram_per_minute = Unit('kg/min', defined_by=kilogram / minute)
kilogram_per_day = Unit('kg/d', defined_by=kilogram / day)
tonne_per_day = Unit('t/d', defined_by=tonne / day)
tonne_per_hour = Unit('t/h', defined_by=tonne / hour)
tonne_per_year = Unit('t/yr', defined_by=tonne / year)
gram_per_hour = Unit('g/h', defined_by=gram / hour)
gram_per_minute = Unit('g/min', defined_by=gram / minute)
gram_per_second = Unit('g/s', defined_by=gram / second)
pound_per_hour = Unit('lb/h', defined_by=pound / hour)
kilo_pound_per_hour = Unit('klb/h', defined_by=1e3 * pound / hour)
pound_per_day = Unit('lb/d', defined_by=pound / day)
kilo_pound_per_day = Unit('klb/d', defined_by=1e3 * pound / day)
million_pound_per_day = Unit('MMlb/d', defined_by=1e6 * pound / day)

# volume flow, base: m3/s
cubic_meter_per_second = Unit('m**3/s')
cubic_meter_per_hour = Unit('m**3/h', defined_by=cubic_meter / hour)
cubic_meter_per_minute = Unit('m**3/min', defined_by=cubic_meter / minute)
cubic_meter_per_day = Unit('m**3/d', defined_by=cubic_meter / day)
liter_per_hour = Unit('L/h', defined_by=liter / hour)
liter_per_day = Unit('L/d', defined_by=liter / day)
liter_per_minute = Unit('L/min', defined_by=liter / minute)
liter_per_second = Unit('L/s', defined_by=liter / second)
milliliter_per_hour = Unit('mL/h', defined_by=milliliter / hour)
milliliter_per_minute = Unit('mL/min', defined_by=milliliter / minute)
milliliter_per_second = Unit('mL/s', defined_by=milliliter / second)
barrel_per_day = Unit('bbl/d', defined_by=barrel / day)
barrel_per_hour = Unit('bbl/h', defined_by=barrel / hour)
million_gallon_per_day = Unit('MMgal/d', defined_by=1e6 * gallon / day)
US_gallon_per_minute = Unit('USGPM', defined_by=gallon / minute)
US_gallon_per_hour = Unit('USGPH', defined_by=gallon / hour)
cubic_foot_per_hour = Unit('ft**3/h', defined_by=cubic_foot / hour)
cubic_foot_per_day = Unit('ft**3/d', defined_by=cubic_foot / day)

# Energy 'kJ/s'
kilojoule_per_second = Unit('kJ/s')
kilojoule_per_hour = Unit('kJ/h', defined_by=kilojoule / hour)
kilojoule_per_minute = Unit('kJ/min', defined_by=kilojoule / minute)
megajoule_per_hour = Unit('MJ/h', defined_by=megajoule / hour)
gigajoule_per_hour = Unit('GJ/h', defined_by=gigajoule / hour)
kilowatt = Unit('kW', defined_by=kilojoule / second)
megawatt = Unit('MW', defined_by=1e3 * kilowatt)
kilocalorie_per_hour = Unit('kcal/h', defined_by=kilocalorie / hour)
kilocalorie_per_minute = Unit('kcal/min', defined_by=kilocalorie / minute)
kilocalorie_per_second = Unit('kcal/s', defined_by=kilocalorie / second)
million_kilocalorie_per_hour = Unit('MMkcal/h', defined_by=million_kilocalorie / hour)
calorie_per_hour = Unit('cal/h', defined_by=calorie / hour)
calorie_per_minute = Unit('cal/min', defined_by=calorie / minute)
calorie_per_second = Unit('cal/s', defined_by=calorie / second)
Btu_per_hour = Unit('Btu/h', defined_by=british_thermal_unit / hour)
million_Btu_per_hour = Unit('MMBtu/h', defined_by=million_british_thermal_unit / hour)
million_Btu_per_day = Unit('MMBtu/d', defined_by=million_british_thermal_unit / day)
horse_power = Unit('hp', defined_by=0.745699 * kilowatt)

# molar density, 'kmol/m3
kilomole_per_cubic_meter = Unit('kmol/m**3')
mole_per_liter = Unit('mol/L', defined_by=mole / liter)
mole_per_cubic_centimeter = Unit('mol/cm**3', defined_by=mole / cubic_centimeter)
mole_per_milliliter = Unit('mol/mL', defined_by=mole / milliliter)

# heat capacity, entropy, 'kJ/kmol-C'
kilojoule_per_mole_celsius = Unit('kJ/(mol*C)', defined_by=kilojoule / (mole * delta_celsius))
kilojoule_per_mole_kelvin = Unit('kJ/(mol*K)', defined_by=kilojoule / (mole * delta_kelvin))
kilojoule_per_kilomole_celsius = Unit('kJ/(kmol*C)')
kilojoule_per_kilomole_kelvin = Unit('kJ/(kmol*K)', defined_by=kilojoule / (kilomole * delta_kelvin))
joule_per_mole_celsius = Unit('J/(mol*C)', defined_by=joule / (mole * delta_celsius))
joule_per_mole_kelvin = Unit('J/(mol*K)', defined_by=joule / (mole * delta_kelvin))
joule_per_kilomole_celsius = Unit('J/(kmol*C)', defined_by=joule / (kilomole * delta_celsius))
joule_per_kilomole_kelvin = Unit('J/(kmol*K)', defined_by=joule / (kilomole * delta_kelvin))
kilocalorie_per_mole_celsius = Unit('kcal/(mol*C)', defined_by=kilocalorie / (mole * delta_celsius))
kilocalorie_per_mole_kelvin = Unit('kcal/(mol*K)', defined_by=kilocalorie / (mole * delta_kelvin))
kilocalorie_per_kilomole_celsius = Unit('kcal/(kmol*C)', defined_by=kilocalorie / (kilomole * delta_celsius))
kilocalorie_per_kilomole_kelvin = Unit('kcal/(kmol*K)', defined_by=kilocalorie / (kilomole * delta_kelvin))
calorie_per_mole_celsius = Unit('cal/(mol*C)', defined_by=calorie / (mole * delta_celsius))
calorie_per_mole_kelvin = Unit('cal/(mol*K)', defined_by=calorie / (mole * delta_kelvin))
calorie_per_kilomole_celsius = Unit('cal/(kmol*C)', defined_by=calorie / (kilomole * delta_celsius))
calorie_per_kilomole_kelvin = Unit('cal/(kmol*K)', defined_by=calorie / (kilomole * delta_kelvin))

# thermal conductivity, W/m-K
watt_per_meter_kelvin = Unit('W/(m*K)')
Btu_per_hour_foot_fahrenheit = Unit('Btu/(h*ft*F)', defined_by=1e3 * british_thermal_unit / (
        hour * foot * delta_fahrenheit))
kilocalorie_per_meter_hour_celsius = Unit('kcal/(m*h*C)', defined_by=1e3 * kilocalorie / hour)
calorie_per_centimeter_second_celsius = Unit('cal/(cm*s*C)', defined_by=1e3 * calorie / centimeter)

# viscosity, cP
centipoise = Unit('cP')
millipoise = Unit('mP', factor=0.1)
micropoise = Unit('microP', defined_by=1e-3 * millipoise)
poise = Unit('P', factor=100)
pascal_second = Unit('Pa-s', factor=1000)
pound_force_second_per_square_foot = Unit('lbf*s/ft**2', defined_by=1e3 * pound_force / square_foot)
pound_mass_per_foot_second = Unit('lbm/(ft*s)', defined_by=1e3 * pound / foot)
pound_mass_per_foot_hour = Unit('lbm/(ft*h)', defined_by=1e3 * pound / (foot * hour))

# surface tension, dyne/cm
dyne_per_centimeter = Unit('dyne/cm')
dyn_per_centimeter = Unit('dyn/cm', factor=1)
pound_force_per_foot = Unit('lbf/ft', defined_by=1e3 * pound_force / foot)

# mass capacity, kJ/kg-C
kilojoule_per_gram_celsius = Unit('kJ/(g*C)', defined_by=kilojoule / (gram * delta_celsius))
kilojoule_per_gram_kelvin = Unit('kJ/(g*K)', defined_by=kilojoule / (gram * delta_kelvin))
kilojoule_per_kilogram_celsius = Unit('kJ/(kg*C)')
kilojoule_per_kilogram_kelvin = Unit('kJ/(kg*K)', defined_by=kilojoule / (kilogram * delta_kelvin))
joule_per_gram_celsius = Unit('J/(g*C)', defined_by=joule / (gram * delta_celsius))
joule_per_gram_kelvin = Unit('J/(g*K)', defined_by=joule / (gram * delta_kelvin))
joule_per_kilogram_celsius = Unit('J/(kg*C)', defined_by=joule / (kilogram * delta_celsius))
joule_per_kilogram_kelvin = Unit('J/(kg*K)', defined_by=joule / (kilogram * delta_kelvin))
kilocalorie_per_gram_celsius = Unit('kcal/(g*C)', defined_by=kilocalorie / (gram * delta_celsius))
kilocalorie_per_gram_kelvin = Unit('kcal/(g*K)', defined_by=kilocalorie / (gram * delta_kelvin))
kilocalorie_per_kilogram_celsius = Unit('kcal/(kg*C)', defined_by=kilocalorie / (kilogram * delta_celsius))
kilocalorie_per_kilogram_kelvin = Unit('kcal/(kg*K)', defined_by=kilocalorie / (kilogram * delta_kelvin))
calorie_per_gram_celsius = Unit('cal/(g*C)', defined_by=calorie / (gram * delta_celsius))
calorie_per_gram_kelvin = Unit('cal/(g*K)', defined_by=calorie / (gram * delta_kelvin))
calorie_per_kilogram_celsius = Unit('cal/(kg*C)', defined_by=calorie / (kilogram * delta_celsius))
calorie_per_kilogram_kelvin = Unit('cal/(kg*K)', defined_by=calorie / (kilogram * delta_kelvin))

# mass density, kg/m3
kilogram_per_cubic_meter = Unit('kg/m**3')
gram_per_liter = Unit('g/L', defined_by=gram / liter)
gram_per_cubic_centimeter = Unit('g/cm**3', defined_by=gram / cubic_centimeter)
gram_per_milliliter = Unit('g/mL', defined_by=gram / milliliter)

# molar_enthalpy
kilojoule_per_mole = Unit('kJ/mol', defined_by=kilojoule / mole)
kilojoule_per_kilomole = Unit('kJ/kmol')
joule_per_mole = Unit('J/mol', defined_by=joule / mole)
joule_per_kilomole = Unit('J/kmol', defined_by=joule / kilomole)
megajoule_per_kilomole = Unit('MJ/kmol', defined_by=megajoule / kilomole)
kilocalorie_per_mole = Unit('kcal/mol', defined_by=kilocalorie / mole)
kilocalorie_per_kilomole = Unit('kcal/kmol', defined_by=kilocalorie / kilomole)
calorie_per_mole = Unit('cal/mol', defined_by=calorie / mole)
calorie_per_kilomole = Unit('cal/kmol', defined_by=calorie / kilomole)

# molar_volume
cubic_meter_per_mole = Unit('m**3/mol', defined_by=cubic_meter / mole)
cubic_meter_per_kilomole = Unit('m**3/kmol')
liter_per_mole = Unit('L/mol', defined_by=liter / mole)
cubic_centimeter_per_mole = Unit('cm**3/mol', defined_by=cubic_centimeter / mole)
milliliter_per_mole = Unit('mL/mol', defined_by=milliliter / mole)

# mass_heat
kilojoule_per_gram = Unit('kJ/g', defined_by=kilojoule / gram)
kilojoule_per_kilogram = Unit('kJ/kg')
joule_per_gram = Unit('J/g', defined_by=joule / gram)
joule_per_kilogram = Unit('J/kg', defined_by=joule / kilogram)
megajoule_per_kilogram = Unit('MJ/kg', defined_by=megajoule / kilogram)
kilocalorie_per_gram = Unit('kcal/g', defined_by=kilocalorie / gram)
kilocalorie_per_kilogram = Unit('kcal/kg', defined_by=kilocalorie / kilogram)
calorie_per_gram = Unit('cal/g', defined_by=calorie / gram)
calorie_per_kilogram = Unit('cal/kg', defined_by=calorie / kilogram)
Btu_per_pound = Unit('Btu/lb', defined_by=british_thermal_unit / pound)

# fraction
dimensionless = Unit('')
percent = Unit('%', factor=1e-2)
parts_per_million = Unit('ppm', factor=1e-6)

# molecular_weight
gram_per_mole = Unit('g/mol', factor=1)
kilogram_per_kilomole = Unit('kg/kmol')
kilogram_per_mole = Unit('kg/mol', factor=1e3)



from z_unit import quantity
from z_unit import unit
from z_unit.quantity import Quantity

quantity_registry = {cls: cls.get_unit_registry().symbols for cls in quantity.Quantity.__subclasses__()}


def convert(value, from_unit: str, to_unit: str) -> Quantity:
    """
    Convert value with unit to target unit
    :param value:
    :param from_unit:
    :param to_unit:
    :return: Quantity instance if units match, otherwise return None.

    >>> convert(1, 'm', 'ft')
    <Length(3.2808399, 'ft')>
    """
    for Q, symbols in quantity_registry.items():
        if {from_unit, to_unit}.issubset(symbols):
            return Q(value, from_unit).to(to_unit)
        
    raise ValueError(f"Cannot convert from '{from_unit}' to '{to_unit}'")



from __future__ import annotations
from typing import Iterable, Optional

from z_unit import unit as u
from z_unit.unit import Unit


class UnitRegistry:
    def __init__(self, name: str):
        self._name = name
        self._units: dict[str, Unit] = {}
        self._base_unit: Optional[Unit] = None
        

    def register(self, unit: Unit, is_base: bool = False):
        """
        Register a unit in the registry.

        Args:
            unit: The unit to register.
            is_base: Whether the unit is the base unit.

        Raises:
            ValueError: If the unit is already registered or if the base unit is already set.
        """
        if unit.symbol in self._units:
            raise ValueError(f"Unit {unit.symbol} already registered")

        unit._registry = self
        self._units[unit.symbol] = unit

        if is_base:
            if self._base_unit is not None:
                raise ValueError(f"Base unit already set to {self._base_unit.symbol}")
            self._base_unit = unit

    @property
    def base_unit(self) -> Optional[Unit]:
        return self._base_unit

    def get_unit(self, symbol: str) -> Unit:
        if symbol not in self._units:
            raise ValueError(f"Unit '{symbol}' not found")
        return self._units[symbol]        

    @property
    def units(self) -> list[Unit]:
        return list(self._units.values())

    @property
    def symbols(self) -> list[str]:
        return list(self._units.keys())


length = UnitRegistry(name="length")
length.register(u.meter, is_base=True)
length.register(u.kilometer)
length.register(u.decimeter)
length.register(u.centimeter)
length.register(u.millimeter)
length.register(u.micrometer)
length.register(u.foot)
length.register(u.inch)

area = UnitRegistry(name="area")
area.register(u.square_meter, is_base=True)
area.register(u.square_kilometer)
area.register(u.square_decimeter)
area.register(u.square_centimeter)
area.register(u.square_millimeter)
area.register(u.square_micrometer)
area.register(u.square_foot)
area.register(u.square_inch)

volume = UnitRegistry(name="volume")
volume.register(u.cubic_meter, is_base=True)
volume.register(u.cubic_centimeter)
volume.register(u.cubic_millimeter)
volume.register(u.millimeter)
volume.register(u.liter)
volume.register(u.milliliter)
volume.register(u.cubic_foot)
volume.register(u.cubic_inch)
volume.register(u.gallon)
volume.register(u.barrel)

time = UnitRegistry(name="time")
time.register(u.second, is_base=True)
time.register(u.minute)
time.register(u.hour)
time.register(u.day)
time.register(u.week)
time.register(u.year)
time.register(u.month)

mass = UnitRegistry(name="mass")
mass.register(u.kilogram, is_base=True)
mass.register(u.gram)
mass.register(u.tonne)
mass.register(u.pound)

force = UnitRegistry(name="force")
force.register(u.newton, is_base=True)
force.register(u.kilogram_meter_per_second_squared)
force.register(u.kilo_newton)
force.register(u.dyne)
force.register(u.kilogram_force)
force.register(u.tonne_force)
force.register(u.pound_force)

substance = UnitRegistry(name="substance")  
substance.register(u.kilomole, is_base=True)
substance.register(u.mole)
substance.register(u.normal_cubic_meter)
substance.register(u.standard_cubic_meter)
substance.register(u.standard_cubic_meter_20C)
substance.register(u.standard_cubic_meter_60F)
substance.register(u.standard_cubic_foot)
substance.register(u.kilo_standard_cubic_foot)
substance.register(u.million_standard_cubic_foot)

energy = UnitRegistry(name="energy")
energy.register(u.kilojoule, is_base=True)
energy.register(u.joule)
energy.register(u.megajoule)
energy.register(u.gigajoule)
energy.register(u.kilowatt_hour)
energy.register(u.kilowatt_year)
energy.register(u.calorie)
energy.register(u.kilocalorie)
energy.register(u.megacalorie)
energy.register(u.gigacalorie)
energy.register(u.million_kilocalorie)
energy.register(u.british_thermal_unit)
energy.register(u.million_british_thermal_unit)
energy.register(u.pound_force_foot)

velocity = UnitRegistry(name="velocity")
velocity.register(u.meter_per_second, is_base=True)
velocity.register(u.meter_per_minute)
velocity.register(u.meter_per_hour)
velocity.register(u.kilometer_per_hour)
velocity.register(u.centimeter_per_second)
velocity.register(u.foot_per_second)
velocity.register(u.foot_per_minute)
velocity.register(u.foot_per_hour)

temperature = UnitRegistry(name="temperature")
temperature.register(u.celsius, is_base=True)
temperature.register(u.kelvin)
temperature.register(u.fahrenheit)
temperature.register(u.rankine)

delta_temperature = UnitRegistry(name="delta_temperature")
delta_temperature.register(u.delta_celsius, is_base=True)
delta_temperature.register(u.delta_kelvin)
delta_temperature.register(u.delta_rankine)
delta_temperature.register(u.delta_fahrenheit)

pressure = UnitRegistry(name="pressure")
pressure.register(u.pascal, is_base=True)
pressure.register(u.kilopascal)
pressure.register(u.megapascal)
pressure.register(u.bar)
pressure.register(u.millibar)
pressure.register(u.atm)
pressure.register(u.kg_force_per_square_centimeter)
pressure.register(u.psi)
pressure.register(u.pound_force_per_square_foot)
pressure.register(u.torr)
pressure.register(u.mm_Hg)
pressure.register(u.inch_Hg)
pressure.register(u.inch_Hg_60F)
pressure.register(u.kilopascal_gauge)
pressure.register(u.megapascal_gauge)
pressure.register(u.bar_gauge)
pressure.register(u.millibar_gauge)
pressure.register(u.kg_force_per_square_centimeter_gauge)
pressure.register(u.psi_gauge)
pressure.register(u.pound_force_per_square_foot_gauge)
pressure.register(u.torr_gauge)
pressure.register(u.mm_Hg_gauge)
pressure.register(u.inch_Hg_gauge)
pressure.register(u.inch_Hg_60F_gauge)

molar_flow = UnitRegistry(name="molar_flow")
molar_flow.register(u.kilomole_per_second, is_base=True)
molar_flow.register(u.kilomole_per_hour)
molar_flow.register(u.kilomole_per_minute)
molar_flow.register(u.normal_cubic_meter_per_hour)
molar_flow.register(u.normal_cubic_meter_per_day)
molar_flow.register(u.standard_cubic_meter_per_hour)
molar_flow.register(u.standard_cubic_meter_20C_per_hour)
molar_flow.register(u.standard_cubic_meter_60F_per_hour)
molar_flow.register(u.standard_cubic_meter_per_day)
molar_flow.register(u.standard_cubic_meter_20C_per_day)
molar_flow.register(u.standard_cubic_meter_60F_per_day)
molar_flow.register(u.mole_per_hour)
molar_flow.register(u.mole_per_minute)
molar_flow.register(u.mole_per_second)
molar_flow.register(u.standard_cubic_foot_per_day)
molar_flow.register(u.kilo_standard_cubic_foot_per_hour)
molar_flow.register(u.kilo_standard_cubic_foot_per_day)
molar_flow.register(u.million_standard_cubic_foot_per_hour)
molar_flow.register(u.million_standard_cubic_foot_per_day)

mass_flow = UnitRegistry(name="mass_flow")
mass_flow.register(u.kilogram_per_hour)
mass_flow.register(u.kilogram_per_minute)
mass_flow.register(u.kilogram_per_second, is_base=True)
mass_flow.register(u.kilogram_per_day)
mass_flow.register(u.tonne_per_day)
mass_flow.register(u.tonne_per_hour)
mass_flow.register(u.tonne_per_year)
mass_flow.register(u.gram_per_hour)
mass_flow.register(u.gram_per_minute)
mass_flow.register(u.gram_per_second)
mass_flow.register(u.pound_per_hour)
mass_flow.register(u.pound_per_day)
mass_flow.register(u.kilo_pound_per_hour)
mass_flow.register(u.kilo_pound_per_day)
mass_flow.register(u.million_pound_per_day)

volume_flow = UnitRegistry(name="volume_flow")
volume_flow.register(u.cubic_meter_per_hour)
volume_flow.register(u.cubic_meter_per_minute)
volume_flow.register(u.cubic_meter_per_second, is_base=True)
volume_flow.register(u.cubic_meter_per_day)
volume_flow.register(u.liter_per_hour)
volume_flow.register(u.liter_per_day)
volume_flow.register(u.liter_per_minute)
volume_flow.register(u.liter_per_second)
volume_flow.register(u.milliliter_per_hour)
volume_flow.register(u.milliliter_per_minute)
volume_flow.register(u.milliliter_per_second)
volume_flow.register(u.barrel_per_day)
volume_flow.register(u.barrel_per_hour)
volume_flow.register(u.million_gallon_per_day)
volume_flow.register(u.US_gallon_per_minute)
volume_flow.register(u.US_gallon_per_hour)
volume_flow.register(u.cubic_foot_per_hour)
volume_flow.register(u.cubic_foot_per_day)

energy_flow = UnitRegistry(name="energy_flow")
energy_flow.register(u.kilojoule_per_hour)
energy_flow.register(u.kilojoule_per_minute)
energy_flow.register(u.kilojoule_per_second, is_base=True)
energy_flow.register(u.megajoule_per_hour)
energy_flow.register(u.gigajoule_per_hour)
energy_flow.register(u.kilowatt)
energy_flow.register(u.megawatt)
energy_flow.register(u.kilocalorie_per_hour)
energy_flow.register(u.kilocalorie_per_minute)
energy_flow.register(u.kilocalorie_per_second)
energy_flow.register(u.million_kilocalorie_per_hour)
energy_flow.register(u.calorie_per_hour)
energy_flow.register(u.calorie_per_minute)
energy_flow.register(u.calorie_per_second)
energy_flow.register(u.Btu_per_hour)
energy_flow.register(u.million_Btu_per_hour)
energy_flow.register(u.million_Btu_per_day)
energy_flow.register(u.horse_power)

heat_flow = energy_flow

molar_density = UnitRegistry(name="molar_density")
molar_density.register(u.kilomole_per_cubic_meter, is_base=True)
molar_density.register(u.mole_per_liter)
molar_density.register(u.mole_per_cubic_centimeter)
molar_density.register(u.mole_per_milliliter)

molar_heat_capacity = UnitRegistry(name="molar_heat_capacity")
molar_heat_capacity.register(u.kilojoule_per_mole_celsius)
molar_heat_capacity.register(u.kilojoule_per_mole_kelvin)
molar_heat_capacity.register(u.kilojoule_per_kilomole_celsius, is_base=True)
molar_heat_capacity.register(u.kilojoule_per_kilomole_kelvin)
molar_heat_capacity.register(u.joule_per_mole_celsius)
molar_heat_capacity.register(u.joule_per_mole_kelvin)
molar_heat_capacity.register(u.joule_per_kilomole_celsius)
molar_heat_capacity.register(u.joule_per_kilomole_kelvin)
molar_heat_capacity.register(u.kilocalorie_per_mole_celsius)
molar_heat_capacity.register(u.kilocalorie_per_mole_kelvin)
molar_heat_capacity.register(u.kilocalorie_per_kilomole_celsius)
molar_heat_capacity.register(u.kilocalorie_per_kilomole_kelvin)
molar_heat_capacity.register(u.calorie_per_mole_celsius)
molar_heat_capacity.register(u.calorie_per_mole_kelvin)
molar_heat_capacity.register(u.calorie_per_kilomole_celsius)
molar_heat_capacity.register(u.calorie_per_kilomole_kelvin)

molar_entropy = molar_heat_capacity

thermal_conductivity = UnitRegistry(name="thermal_conductivity")
thermal_conductivity.register(u.watt_per_meter_kelvin, is_base=True)
thermal_conductivity.register(u.Btu_per_hour_foot_fahrenheit)
thermal_conductivity.register(u.kilocalorie_per_meter_hour_celsius)
thermal_conductivity.register(u.calorie_per_centimeter_second_celsius)

viscosity = UnitRegistry(name="viscosity")
viscosity.register(u.centipoise, is_base=True)
viscosity.register(u.millipoise)
viscosity.register(u.micropoise)
viscosity.register(u.poise)
viscosity.register(u.pascal_second)
viscosity.register(u.pound_force_second_per_square_foot)
viscosity.register(u.pound_mass_per_foot_second)
viscosity.register(u.pound_mass_per_foot_hour)

surface_tension = UnitRegistry(name="surface_tension")
surface_tension.register(u.dyne_per_centimeter, is_base=True)
surface_tension.register(u.dyn_per_centimeter)
surface_tension.register(u.pound_force_per_foot)

mass_heat_capacity = UnitRegistry(name="mass_heat_capacity")    
mass_heat_capacity.register(u.kilojoule_per_gram_celsius)
mass_heat_capacity.register(u.kilojoule_per_gram_kelvin)
mass_heat_capacity.register(u.kilojoule_per_kilogram_celsius, is_base=True)
mass_heat_capacity.register(u.kilojoule_per_kilogram_kelvin)
mass_heat_capacity.register(u.joule_per_gram_celsius)
mass_heat_capacity.register(u.joule_per_gram_kelvin)
mass_heat_capacity.register(u.joule_per_kilogram_celsius)
mass_heat_capacity.register(u.joule_per_kilogram_kelvin)
mass_heat_capacity.register(u.kilocalorie_per_gram_celsius)
mass_heat_capacity.register(u.kilocalorie_per_gram_kelvin)
mass_heat_capacity.register(u.kilocalorie_per_kilogram_celsius)
mass_heat_capacity.register(u.kilocalorie_per_kilogram_kelvin)
mass_heat_capacity.register(u.calorie_per_gram_celsius)
mass_heat_capacity.register(u.calorie_per_gram_kelvin)
mass_heat_capacity.register(u.calorie_per_kilogram_celsius)
mass_heat_capacity.register(u.calorie_per_kilogram_kelvin)

mass_density = UnitRegistry(name="mass_density")
mass_density.register(u.kilogram_per_cubic_meter, is_base=True)
mass_density.register(u.gram_per_liter)
mass_density.register(u.gram_per_cubic_centimeter)
mass_density.register(u.gram_per_milliliter)

standard_gas_flow = UnitRegistry(name="standard_gas_flow")
standard_gas_flow.register(u.Unit('Sm**3/h', factor=1 / u.hour.factor))
standard_gas_flow.register(u.Unit('Sm**3/d', factor=1 / u.day.factor))
standard_gas_flow.register(u.Unit('Sm**3/min', factor=1 / u.minute.factor))
standard_gas_flow.register(u.Unit('Sm**3/s'), is_base=True)

molar_enthalpy = UnitRegistry(name="molar_enthalpy")
molar_enthalpy.register(u.kilojoule_per_mole)
molar_enthalpy.register(u.kilojoule_per_kilomole, is_base=True)
molar_enthalpy.register(u.joule_per_mole)
molar_enthalpy.register(u.joule_per_kilomole)
molar_enthalpy.register(u.megajoule_per_kilomole)
molar_enthalpy.register(u.kilocalorie_per_mole)
molar_enthalpy.register(u.kilocalorie_per_kilomole)
molar_enthalpy.register(u.calorie_per_mole)
molar_enthalpy.register(u.calorie_per_kilomole)

molar_heat = molar_enthalpy

molar_energy = molar_heat

molar_volume = UnitRegistry(name="molar_volume")
molar_volume.register(u.cubic_meter_per_mole)
molar_volume.register(u.cubic_meter_per_kilomole, is_base=True)
molar_volume.register(u.liter_per_mole)
molar_volume.register(u.cubic_centimeter_per_mole)
molar_volume.register(u.milliliter_per_mole)

mass_entropy = mass_heat_capacity

mass_heat = UnitRegistry(name="mass_heat")
mass_heat.register(u.kilojoule_per_gram)
mass_heat.register(u.kilojoule_per_kilogram, is_base=True)
mass_heat.register(u.joule_per_gram)
mass_heat.register(u.joule_per_kilogram)
mass_heat.register(u.megajoule_per_kilogram)
mass_heat.register(u.kilocalorie_per_gram)
mass_heat.register(u.kilocalorie_per_kilogram)
mass_heat.register(u.calorie_per_gram)
mass_heat.register(u.calorie_per_kilogram)
mass_heat.register(u.Btu_per_pound)

mass_enthalpy = mass_heat

mass_energy = mass_heat

kinematic_viscosity = UnitRegistry(name="kinematic_viscosity")
kinematic_viscosity.register(u.Unit('cSt'), is_base=True)

fraction = UnitRegistry(name="fraction")
fraction.register(u.dimensionless, is_base=True)
fraction.register(u.percent)
fraction.register(u.parts_per_million)

dimensionless = UnitRegistry(name="dimensionless")
dimensionless.register(u.dimensionless, is_base=True)



from math import isclose

from z_unit.environment import get_env
from z_unit import quantity as q


def test_convert():
    env = get_env()
    env.reset()
    env.standard_temperature = 293.15
    f = q.MolarFlow(1, 'Nm3/h')
    f1 = f.to('Sm3/h').value
    env.standard_temperature = 273.15
    f2 = f.to('Sm3/h').value
    assert not isclose(f1, f2)
    env.reset()



from math import isclose

from z_unit import quantity as q
from z_unit.environment import get_env


def test_length():
    x = q.Length(1)
    assert x == q.Length('1m')
    assert x == q.Length('1')
    assert x == q.Length('1 m')
    assert x.unit.symbol == 'm'
    assert x.unit == x.base_unit
    assert f'{x:u}' == '1 m'
    assert f'{x:uq}' == '1 m'
    assert f'{x:up}' == '1 m'
    print(f"Length(1) * 5 = {x * 5:u}")
    print(f"5 * Length(1) = {5 * x:u}")
    print(f"5 * Length(1, 'ft') = {5 * q.Length(1, 'ft'):u}")
    assert isclose(x.to('km').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('dm').value, 1e1, rel_tol=1e-4)
    assert isclose(x.to('cm').value, 1e2, rel_tol=1e-4)
    assert isclose(x.to('mm').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('um').value, 1e6, rel_tol=1e-4)
    assert isclose(x.to('ft').value, 3.28084, rel_tol=1e-4)
    assert isclose(x.to('in').value, 39.37008, rel_tol=1e-4)


def test_area():
    x = q.Area(1)
    assert x.unit.symbol == 'm2'
    assert f'{x:u}' == '1 m2'
    assert f'{x:up}' == '1 m**2'
    assert x.unit == x.base_unit
    assert isclose(x.to('km2').value, 1e-6, rel_tol=1e-4)
    assert isclose(x.to('dm2').value, 1e2, rel_tol=1e-4)
    assert isclose(x.to('cm2').value, 1e4, rel_tol=1e-4)
    assert isclose(x.to('mm2').value, 1e6, rel_tol=1e-4)
    assert isclose(x.to('um2').value, 1e12, rel_tol=1e-4)
    assert isclose(x.to('ft2').value, 10.7639, rel_tol=1e-4)
    assert isclose(x.to('in2').value, 1550.0016000682, rel_tol=1e-4)


def test_volume():
    x = q.Volume(1)
    assert x.unit.symbol == 'm3'
    assert x.unit == x.base_unit
    assert isclose(x.to('cm3').value, 1e6, rel_tol=1e-4)
    assert isclose(x.to('mm3').value, 1e9, rel_tol=1e-4)
    assert isclose(x.to('L').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('mL').value, 1e6, rel_tol=1e-4)
    assert isclose(x.to('ft3').value, 35.3147, rel_tol=1e-4)
    assert isclose(x.to('in3').value, 61023.801579099, rel_tol=1e-4)
    assert isclose(x.to('gal').value, 264.1723012, rel_tol=1e-4)
    assert isclose(x.to('bbl').value, 6.28981, rel_tol=1e-4)


def test_time():
    x = q.Time(1)
    assert x.unit.symbol == 's'
    assert x.unit == x.base_unit
    assert isclose(x.to('min').value, 1 / 60, rel_tol=1e-4)
    assert isclose(x.to('hr').value, 1 / 3600, rel_tol=1e-4)
    assert isclose(x.to('day').value, 1 / 3600 / 24, rel_tol=1e-4)
    assert isclose(x.to('yr').value, 1 / 3600 / 8760, rel_tol=1e-4)
    assert isclose(x.to('mon').value, 1 / 3600 / 8760 * 12, rel_tol=1e-4)


def test_mass():
    x = q.Mass(1)
    assert x.unit.symbol == 'kg'
    assert x.unit == x.base_unit
    assert isclose(x.to('g').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('t').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('lb').value, 2.20462, rel_tol=1e-4)


def test_force():
    x = q.Force(1)
    assert x.unit.symbol == 'N'
    assert x.unit == x.base_unit
    assert isclose(x.to('kg-m/s2').value, 1, rel_tol=1e-4)
    assert isclose(x.to('kN').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('dyn').value, 1e5, rel_tol=1e-4)
    assert isclose(x.to('kgf').value, 0.101972, rel_tol=1e-4)
    assert isclose(x.to('tonf').value, 0.101972e-3, rel_tol=1e-4)
    assert isclose(x.to('lbf').value, 0.224809, rel_tol=1e-4)


def test_substance():
    x = q.Substance(1)
    assert x.unit.symbol == 'kmol'
    assert x.unit == x.base_unit
    assert isclose(x.to('mol').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('Nm3').value, 22.414, rel_tol=1e-4)
    assert isclose(x.to('Sm3_20C').value, 24.055149, rel_tol=1e-4)
    assert isclose(x.to('SCF').value, 836.56108, rel_tol=1e-4)
    assert isclose(x.to('MSCF').value, 836.56108e-3, rel_tol=1e-4)
    assert isclose(x.to('MMSCF').value, 836.56108e-6, rel_tol=1e-4)
    env = get_env()
    env.standard_temperature = 273.15
    assert isclose(x.to('Sm3').value, x.to('Nm3').value)
    env.reset()


def test_energy():
    x = q.Energy(1)
    assert x.unit.symbol == 'kJ'
    assert x.unit == x.base_unit
    assert isclose(x.to('J').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('MJ').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('GJ').value, 1e-6, rel_tol=1e-4)
    assert isclose(x.to('kW-h').value, 0.000277778, rel_tol=1e-4)
    assert isclose(x.to('kW-yr').value, 1 / 3600 / 8760, rel_tol=1e-4)
    assert isclose(x.to('cal').value, 0.239006e3, rel_tol=1e-4)
    assert isclose(x.to('kcal').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('Mcal').value, 0.239006e-3, rel_tol=1e-4)
    assert isclose(x.to('Gcal').value, 0.239006e-6, rel_tol=1e-4)
    assert isclose(x.to('MMkcal').value, 0.239006e-6, rel_tol=1e-4)
    assert isclose(x.to('Btu').value, 0.947818, rel_tol=1e-4)
    assert isclose(x.to('MMBtu').value, 0.947818e-6, rel_tol=1e-4)
    assert isclose(x.to('lbf-ft').value, 737.562, rel_tol=1e-4)


def test_velocity():
    x = q.Velocity(1)
    assert x.unit.symbol == 'm/s'
    assert x.unit == x.base_unit
    assert isclose(x.to('m/min').value, 60, rel_tol=1e-4)
    assert isclose(x.to('m/hr').value, 3600, rel_tol=1e-4)
    assert isclose(x.to('km/hr').value, 1e-3 * 3600, rel_tol=1e-4)
    assert isclose(x.to('cm/s').value, 100, rel_tol=1e-4)
    assert isclose(x.to('ft/s').value, 3.28084, rel_tol=1e-4)
    assert isclose(x.to('ft/min').value, 3.28084 * 60, rel_tol=1e-4)
    assert isclose(x.to('ft/hr').value, 3.28084 * 3600, rel_tol=1e-4)


def test_temperature():
    x = q.Temperature(25)
    assert x.value == 25
    assert x.unit.symbol == 'C'
    assert isclose(x.to('K').value, 298.15)
    assert isclose(x.to('F').value, 77)
    assert isclose(x.to('R').value, 536.67)
    assert isclose(q.Temperature(300, 'K').to_base().value, 26.85)
    assert isclose(q.Temperature(200, 'C').to_base().value, 200)
    assert isclose(q.Temperature(100, 'F').to_base().value, 37.7777778)
    assert isclose(q.Temperature(50, 'R').to_base().value, -245.3722222)
    assert isclose(q.Temperature(300, 'K').to('F').value, 80.33)


def test_delta_temperature():
    x = q.DeltaTemperature(1)
    assert x.unit.symbol == 'C'
    assert x.unit == x.base_unit
    assert isclose(x.to('K').value, 1, rel_tol=1e-4)
    assert isclose(x.to('R').value, 1.8, rel_tol=1e-4)
    assert isclose(x.to('F').value, 1.8, rel_tol=1e-4)


def test_pressure():
    x = q.Pressure(100, 'kPa')
    assert x.value == 100
    assert x.unit.symbol == 'kPa'
    assert isclose(x.to('Pa').value, 100e3)
    assert isclose(x.to('MPa').value, 100e-3)
    assert isclose(x.to('bar').value, 100e-2)
    assert isclose(x.to('mbar').value, 1000)
    assert isclose(x.to('kgf/cm2').value, 1.019716, rel_tol=1e-4)
    assert isclose(x.to('atm').value, 0.9869233, rel_tol=1e-4)
    assert isclose(x.to('psi').value, 14.503774, rel_tol=1e-4)
    assert isclose(x.to('lbf/ft2').value, 2088.543, rel_tol=1e-4)
    assert isclose(x.to('torr').value, 750.06168, rel_tol=1e-4)
    assert isclose(x.to('mmHg_0C').value, 750.06168, rel_tol=1e-4)
    assert isclose(x.to('inHg_32F').value, 29.52998, rel_tol=1e-4)
    assert isclose(x.to('inHg_60F').value, 29.61340, rel_tol=1e-4)
    assert isclose(x.to('kPag').value, -1.325, rel_tol=1e-4)
    assert isclose(x.to('MPag').value * 1000, -1.325, rel_tol=1e-4)
    assert isclose(x.to('barg').value * 100, -1.325, rel_tol=1e-4)
    assert isclose(x.to('kgf/cm2_g').value, -0.0135112, rel_tol=1e-4)
    assert isclose(x.to('psig').value, -0.19218, rel_tol=1e-4)
    assert isclose(x.to('lbf/ft2_g').value, -27.673, rel_tol=1e-4)
    assert isclose(x.to('torr_g').value, -9.9385, rel_tol=1e-4)
    assert isclose(x.to('mmHg_0C_g').value, -9.9385, rel_tol=1e-4)
    assert isclose(x.to('inHg_32F_g').value, -0.39127, rel_tol=1e-4)
    assert isclose(x.to('inHg_60F_g').value, -0.39237, rel_tol=1e-4)
    assert isclose(q.Pressure(1, 'MPag').to('psi').value, 159.73368651, rel_tol=1e-4)
    # change atmospheric pressure
    env = get_env()
    env.atmospheric_pressure = 50e3
    assert isclose(x.to('kPag').value, 50)
    # restore atmospheric pressure
    env.reset()
    assert isclose(x.to('kPag').value, -1.325, rel_tol=1e-4)


def test_volume_flow():
    v = q.VolumeFlow(1)
    assert v.unit.symbol == 'm3/s'
    assert v.base_unit.symbol == v.unit.symbol
    assert isclose(v.to('m3/h').value, 3600, rel_tol=1e-6)
    assert isclose(v.to('m3/min').value, 60, rel_tol=1e-6)
    assert isclose(v.to('m3/d').value, 3600 * 24, rel_tol=1e-6)
    assert isclose(v.to('L/h').value, 3600000, rel_tol=1e-6)
    assert isclose(v.to('L/d').value, 3600000 * 24, rel_tol=1e-6)
    assert isclose(v.to('L/min').value, 60000, rel_tol=1e-6)
    assert isclose(v.to('L/s').value, 1000, rel_tol=1e-6)
    assert isclose(v.to('mL/h').value, 3600000000, rel_tol=1e-6)
    assert isclose(v.to('mL/min').value, 60e6, rel_tol=1e-6)
    assert isclose(v.to('mL/s').value, 1e6, rel_tol=1e-6)
    assert isclose(v.to('bbl/d').value, 6.2898106 * 24 * 3600, rel_tol=1e-6)
    assert isclose(v.to('bbl/h').value, 6.2898106 * 3600, rel_tol=1e-6)
    assert isclose(v.to('MMgal/d').value, 264.172 * 24 * 3600 / 1e6, rel_tol=1e-6)
    assert isclose(v.to('USGPM').value, 264.172 * 60, rel_tol=1e-6)
    assert isclose(v.to('USGPH').value, 264.172 * 3600, rel_tol=1e-6)
    assert isclose(v.to('ft3/h').value, 35.3147 * 3600, rel_tol=1e-6)
    assert isclose(v.to('ft3/d').value, 35.3147 * 24 * 3600, rel_tol=1e-6)


def test_mass_density():
    x = q.MassDensity(100)
    assert x.unit.symbol == 'kg/m3'
    assert x.unit == x.base_unit
    assert isclose(x.to('g/L').value, 100, rel_tol=1e-4)
    assert isclose(x.to('g/cm3').value, 0.1, rel_tol=1e-4)
    assert isclose(x.to('g/mL').value, 0.1, rel_tol=1e-4)


def test_heat_flow():
    h = q.HeatFlow(1)
    assert h.unit.symbol == 'kJ/s'
    assert h.unit == h.base_unit
    assert isclose(h.to('kJ/h').value, 3600, rel_tol=1e-6)
    assert isclose(h.to('kJ/min').value, 60, rel_tol=1e-6)
    assert isclose(h.to('MJ/h').value, 3.6, rel_tol=1e-6)
    assert isclose(h.to('GJ/h').value, 3.6e-3, rel_tol=1e-6)
    assert isclose(h.to('kW').value, 1, rel_tol=1e-6)
    assert isclose(h.to('MW').value, 1e-3, rel_tol=1e-6)
    assert isclose(h.to('kcal/h').value, 860.421, rel_tol=1e-6)
    assert isclose(h.to('kcal/min').value, 14.34, rel_tol=1e-4)
    assert isclose(h.to('kcal/s').value, 0.239006, rel_tol=1e-4)
    assert isclose(h.to('MMkcal/h').value, 860.421e-6, rel_tol=1e-4)
    assert isclose(h.to('cal/h').value, 860.421e3, rel_tol=1e-4)
    assert isclose(h.to('cal/min').value, 14.34e3, rel_tol=1e-4)
    assert isclose(h.to('cal/s').value, 0.239006e3, rel_tol=1e-4)
    assert isclose(h.to('Btu/h').value, 3412.14, rel_tol=1e-4)
    assert isclose(h.to('MMBtu/h').value, 3412.14e-6, rel_tol=1e-4)
    assert isclose(h.to('MMBtu/d').value, 3412.14e-6 * 24, rel_tol=1e-4)
    assert isclose(h.to('hp').value, 1.341022, rel_tol=1e-4)


def test_molar_flow():
    x = q.MolarFlow(1)
    assert x.unit.symbol == 'kmol/s'
    assert x.unit == x.base_unit
    assert isclose(x.to('kmol/h').value, 3600, rel_tol=1e-4)
    assert isclose(x.to('kmol/min').value, 60, rel_tol=1e-4)
    assert isclose(x.to('Nm3/h').value, 80690.4, rel_tol=1e-4)
    assert isclose(x.to('Nm3/d').value, 1936569.6, rel_tol=1e-4)
    env = get_env()
    env.standard_temperature = 293.15
    assert isclose(x.to('Sm3/h').value, 86598.538, rel_tol=1e-4)
    assert isclose(x.to('Sm3/d').value, 2078364.92, rel_tol=1e-4)
    assert isclose(x.to('mol/h').value, 3600e3, rel_tol=1e-4)
    assert isclose(x.to('mol/min').value, 60e3, rel_tol=1e-4)
    assert isclose(x.to('mol/s').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('SCFD').value, 7.22788776e7, rel_tol=1e-4)
    assert isclose(x.to('MSCFH').value, 3.0116196e3, rel_tol=1e-4)
    assert isclose(x.to('MSCFD').value, 7.22788776e4, rel_tol=1e-4)
    assert isclose(x.to('MMSCFH').value, 3.0116196, rel_tol=1e-4)
    assert isclose(x.to('MMSCFD').value, 72.2788776, rel_tol=1e-4)
    env.reset()
    

def test_mass_flow():
    x = q.MassFlow(1)
    assert x.unit.symbol == 'kg/s'
    assert x.unit == x.base_unit
    assert isclose(x.to('kg/h').value, 3600, rel_tol=1e-4)
    assert isclose(x.to('kg/min').value, 60, rel_tol=1e-4)
    assert isclose(x.to('kg/d').value, 86400, rel_tol=1e-4)
    assert isclose(x.to('t/d').value, 86.4, rel_tol=1e-4)
    assert isclose(x.to('t/h').value, 3.6, rel_tol=1e-4)
    assert isclose(x.to('t/yr').value, 31536, rel_tol=1e-4)
    assert isclose(x.to('g/h').value, 3600e3, rel_tol=1e-4)
    assert isclose(x.to('g/min').value, 60e3, rel_tol=1e-4)
    assert isclose(x.to('g/s').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('lb/h').value, 7936.56, rel_tol=1e-4)
    assert isclose(x.to('lb/d').value, 190479.348, rel_tol=1e-4)
    assert isclose(x.to('klb/h').value, 7.93656, rel_tol=1e-4)
    assert isclose(x.to('klb/d').value, 190.47744, rel_tol=1e-4)
    assert isclose(x.to('MMlb/d').value, 0.19047744, rel_tol=1e-4)


def test_molar_density():
    x = q.MolarDensity(1)
    assert x.unit.symbol == 'kmol/m3'
    assert x.unit == x.base_unit
    assert isclose(x.to('mol/L').value, 1, rel_tol=1e-4)
    assert isclose(x.to('mol/cm3').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('mol/mL').value, 1e-3, rel_tol=1e-4)


def test_molar_heat_capacity():
    x = q.MolarHeatCapacity(1)
    assert x.unit.symbol == 'kJ/kmol-C'
    assert x.unit == x.base_unit
    assert isclose(x.to('kJ/mol-C').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kJ/mol-K').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kJ/kmol-K').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/mol-C').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/mol-K').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/kmol-C').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('J/kmol-K').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('kcal/kmol-C').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('kcal/kmol-K').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/mol-C').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/mol-K').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/kmol-C').value, 0.239006e3, rel_tol=1e-4)
    assert isclose(x.to('cal/kmol-K').value, 0.239006e3, rel_tol=1e-4)


def test_molar_entropy():
    x = q.MolarEntropy(1)
    assert x.unit.symbol == 'kJ/kmol-C'


def test_molar_heat():
    x = q.MolarHeat(1)
    assert x.unit.symbol == 'kJ/kmol'
    assert x.unit == x.base_unit
    assert isclose(x.to('kJ/mol').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('J/mol').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/kmol').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('MJ/kmol').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/mol').value, 0.239006e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/kmol').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/mol').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/kmol').value, 0.239006e3, rel_tol=1e-4)


def test_thermal_conductivity():
    x = q.ThermalConductivity(1)
    assert x.unit.symbol == 'W/m-K'
    assert x.unit == x.base_unit
    assert isclose(x.to('Btu/h-ft-F').value, 0.577789, rel_tol=1e-4)
    assert isclose(x.to('kcal/m-h-C').value, 0.8604216, rel_tol=1e-4)
    assert isclose(x.to('cal/cm-s-C').value, 0.00239, rel_tol=1e-4)


def test_viscosity():
    x = q.Viscosity(1)
    assert x.unit.symbol == 'cP'
    assert x.unit == x.base_unit
    assert isclose(x.to('mP').value, 10, rel_tol=1e-4)
    assert isclose(x.to('microP').value, 1e4, rel_tol=1e-4)
    assert isclose(x.to('P').value, 1e-2, rel_tol=1e-4)
    assert isclose(x.to('Pa-s').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('lbf-s/ft2').value, 0.0000208854, rel_tol=1e-4)
    assert isclose(x.to('lbm/ft-s').value, 0.000671968994813, rel_tol=1e-4)
    assert isclose(x.to('lbm/ft-h').value, 2.4190883105, rel_tol=1e-4)


def test_surface_tension():
    x = q.SurfaceTension(1)
    assert x.unit.symbol == 'dyne/cm'
    assert x.unit == x.base_unit
    assert isclose(x.to('dyn/cm').value, 1, rel_tol=1e-4)
    assert isclose(x.to('lbf/ft').value, 6.8522e-5, rel_tol=1e-4)


def test_mass_heat_capacity():
    x = q.MassHeatCapacity(1)
    assert x.unit.symbol == 'kJ/kg-C'
    assert x.unit == x.base_unit
    assert isclose(x.to('kJ/g-C').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kJ/g-K').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kJ/kg-K').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/g-C').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/g-K').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/kg-C').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('J/kg-K').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('kcal/g-C').value, 0.239006e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/g-K').value, 0.239006e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/kg-C').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('kcal/kg-K').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/g-C').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/g-K').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/kg-C').value, 0.239006e3, rel_tol=1e-4)
    assert isclose(x.to('cal/kg-K').value, 0.239006e3, rel_tol=1e-4)


def test_mass_entropy():
    x = q.MassEntropy(1)
    assert x.unit.symbol == 'kJ/kg-C'


def test_mass_heat():
    x = q.MassHeat(1)
    assert x.unit.symbol == 'kJ/kg'
    assert x.unit == x.base_unit
    assert isclose(x.to('kJ/g').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('J/g').value, 1, rel_tol=1e-4)
    assert isclose(x.to('J/kg').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('MJ/kg').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/g').value, 0.239006e-3, rel_tol=1e-4)
    assert isclose(x.to('kcal/kg').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/g').value, 0.239006, rel_tol=1e-4)
    assert isclose(x.to('cal/kg').value, 0.239006e3, rel_tol=1e-4)
    assert isclose(x.to('Btu/lb').value, 0.429923, rel_tol=1e-4)


def test_standard_gas_flow():
    x = q.StandardGasFlow(1)
    assert x.unit.symbol == 'Sm3/s'
    assert x.unit == x.base_unit
    assert isclose(x.to('Sm3/h').value, 3600, rel_tol=1e-4)
    assert isclose(x.to('Sm3/d').value, 3600 * 24, rel_tol=1e-4)
    assert isclose(x.to('Sm3/min').value, 60, rel_tol=1e-4)


def test_kinematic_viscosity():
    x = q.KinematicViscosity(1)
    assert x.unit.symbol == 'cSt'
    assert x.unit == x.base_unit


def test_molar_volume():
    x = q.MolarVolume(1)
    assert x.unit.symbol == 'm3/kmol'
    assert x.unit == x.base_unit
    assert isclose(x.to('m3/mol').value, 1e-3, rel_tol=1e-4)
    assert isclose(x.to('L/mol').value, 1, rel_tol=1e-4)
    assert isclose(x.to('mL/mol').value, 1e3, rel_tol=1e-4)
    assert isclose(x.to('cm3/mol').value, 1e3, rel_tol=1e-4)


def test_fraction():
    x = q.Fraction(1)
    assert x.unit.symbol == ''
    assert x.unit == x.base_unit
    assert isclose(x.to('%').value, 100, rel_tol=1e-4)
    assert isclose(x.to('ppm').value, 1e6, rel_tol=1e-4)


def test_dimensionless():
    x = q.Dimensionless(1)
    assert x.unit.symbol == ''
    assert x.unit == x.base_unit


def test_operation():
    x = q.Length(1)
    assert x == q.Length(1)
    assert x != q.Length(2)
    assert x < q.Length(2)
    assert x <= q.Length(1)
    assert x == q.Length(1000, 'mm')
    assert q.Length(100, 'cm') == q.Length(1000, 'mm')
    assert q.Length(200, 'cm') != q.Length(1000, 'mm')
    assert q.Length(200, 'cm') >= q.Length(1000, 'mm')
    assert q.Length(1, 'cm') * 100 == q.Length(1, 'm')
    assert 200 * q.Length(1, 'mm') == q.Length(2, 'dm')
